import os

import machinable as ml
from machinable import Engine
from machinable.config.interface import ConfigInterface
from machinable.experiment.parser import parse_experiment
from machinable.project import Project


def test_experiment_serialization():
    t = (
        ml.Experiment()
        .components(
            "test",
            [
                ("test", [{"a": i} for i in range(3)]),
                ("test", [{"a": i} for i in range(3)]),
            ],
        )
        .repeat(2)
    )
    json = t.to_json()
    t_ = ml.Experiment.from_json(json)
    assert str(t.specification) == str(t_.specification)


def test_experiment_config():
    test_project = Project("./test_project")
    config = test_project.parse_config()

    t = ml.Experiment().components(
        ("nodes.observations", {"attr": "node"}), ("workers.interactive", {"id": 2})
    )
    node, components, resources = list(parse_experiment(t.specification))[0]
    conf = ConfigInterface(config, t.specification["version"])
    node_config = conf.get(node)["args"]
    worker_component_config = conf.get(components[0])["args"]

    assert node_config["attr"] == "node"
    assert worker_component_config["attr"] == "worker"

    t = (
        ml.Experiment()
        .components(
            ("nodes.observations", {"attr": "node"}), ("workers.interactive", {"id": 2})
        )
        .version("~test")
    )
    node, components, resources = list(parse_experiment(t.specification))[0]
    conf = ConfigInterface(config, t.specification["version"])
    node_config = conf.get(node)["args"]
    assert node_config["version"] == 0
    worker_component_config = conf.get(components[0])["args"]
    assert worker_component_config["version"] == 1



def storage_directory_interface():
    _setup(debug=True)
    o = ObservationDirectory("./v1_observations/test_data/tttttt")
    assert o._uid is None
    assert o._path == "tttttt"
    assert o._url == "osfs://./v1_observations/test_data/tttttt"
    o = ObservationDirectory("./v1_observations/test_data/tttttt/tbAXUwxGJzA8")
    assert o._uid == "tbAXUwxGJzA8"
    assert o._path == "tttttt"
    assert o._url == "osfs://./v1_observations/test_data/tttttt"
    assert isinstance(o.load_file("experiment.json", meta=True), dict)



import pytest

from machinable.execution.identifiers import (
    decode_experiment_id,
    encode_experiment_id,
    generate_experiment_id,
)
from machinable.utils.utils import generate_uid


def test_experiment_id_encoding():
    assert encode_experiment_id(946416180) == "123456"
    with pytest.raises(ValueError):
        encode_experiment_id(0)
    assert decode_experiment_id("123456") == 946416180
    with pytest.raises(ValueError):
        decode_experiment_id("invalid")
    for _ in range(5):
        t = generate_experiment_id(with_encoding=False)
        assert decode_experiment_id(encode_experiment_id(t)) == t


def test_uid_generator():
    for seed in [123, 12, 100, 2000]:
        # correct length
        assert len(generate_uid(random_state=seed)[0]) == 12
        assert len(generate_uid(k=5, random_state=seed)) == 5

        # regenerate
        L1 = generate_uid(k=5, random_state=seed)
        L2 = generate_uid(k=5, random_state=seed)

        assert len(L1) == len(L2) and sorted(L1) == sorted(L2)



import base64

from starlette.responses import PlainTextResponse

from ..storage.directory import Directory


async def storage_resolver(request):
    url = base64.urlsafe_b64decode(request.path_params["url"]).decode("utf-8")
    filename = request.path_params["filename"]

    data = Directory(url).load_file(filename, uid="")

    return PlainTextResponse(data)



import importlib
import inspect
import traceback

import regex

from ..utils.dicts import get_or_fail, read_path_dict, update_dict
from ..utils.formatting import msg
from ..utils.utils import is_valid_variable_name
from .mapping import _reserved_keys, _used_keys


class ModuleClass(object):
    def __init__(self, module_name, args=None, baseclass=None):
        self.module_name = module_name
        self.args = args
        self.baseclass = baseclass

    def load(self, instantiate=True, default=None):
        module_class = default
        try:
            module = importlib.import_module(self.module_name)
            importlib.reload(
                module
            )  # reload, in case we are in interactive environments like jupyter

            for candidate, class_ in inspect.getmembers(module, inspect.isclass):
                if self.baseclass is not None and not issubclass(
                    class_, self.baseclass
                ):
                    continue

                if class_.__module__ == self.module_name:
                    module_class = class_
                    break

            if module_class is None:
                raise AttributeError(
                    f"Could not load module class from module '{self.module_name}'. "
                    f"Make sure the module contains a class that inherits from "
                    f"the baseclass 'machinable.{self.baseclass.__name__}'"
                )
        except ImportError as e:
            if default is None:
                trace = "".join(
                    traceback.format_exception(
                        etype=type(e), value=e, tb=e.__traceback__
                    )
                )
                raise ImportError(
                    f"Could not import module '{self.module_name}' "
                    f"that is specified in the machinable.yaml. "
                    f"The following exception occurred: {e}\n{trace}. "
                    f"If the module is a directory, consider creating an __init__.py."
                )

        if not instantiate:
            return module_class

        if self.args is None:
            return module_class()

        return module_class(**self.args)

    def __call__(self, *args, **kwargs):
        class_ = self.load(instantiate=False)
        return class_(*args, **kwargs)

    def __repr__(self):
        return f"<machinable.ModuleClass(module_name={self.module_name}, args={self.args} baseclass={self.baseclass})>"


def parse_mixins(config, valid_only=False):
    if config is None:
        return []

    if isinstance(config, str):
        config = [config]

    if not isinstance(config, (tuple, list)):
        raise ValueError(f"_mixins_ has to be a list. '{config}' given.")

    mixins = []
    for mixin in config:
        if isinstance(mixin, str):
            mixin = dict(name=mixin)

        if "name" not in mixin:
            raise ValueError(f"Mixin definition '{mixin}' must specify a name")

        if "attribute" not in mixin:
            mixin["attribute"] = (
                "_" + mixin["name"].replace("+.", "").replace(".", "_") + "_"
            )

        mixin["valid_attribute"] = is_valid_variable_name(mixin["attribute"])

        if valid_only and not mixin["valid_attribute"]:
            continue

        mixins.append(mixin)

    return mixins


def parse_reference(reference, root, this, wrapped=False):
    result = regex.sub(
        r"""
    (?<rec> #capturing group rec
     \[ #open parenthesis
     (?: #non-capturing group
      [^\[\]]++ #anyting but parenthesis one or more times without backtracking
      | #or
       (?&rec) #recursive substitute of group rec
     )*
     \] #close parenthesis
    )
    """,
        lambda m: parse_reference(m.group()[1:-1], root, this, wrapped=True),
        reference,
        flags=regex.VERBOSE,
    )

    if result.startswith("$."):
        try:
            r = read_path_dict(root, result[2:])
        except KeyError:
            raise KeyError(f"Could not resolve global reference {result}")
    elif result.startswith("$self."):
        try:
            r = read_path_dict(this, result[6:])
        except KeyError:
            raise KeyError(f"Could not resolve local reference {result}")
    else:
        r = result

    if wrapped:
        return "[" + str(r) + "]"

    return r


def parse_references(config, root=None, this=None, validate=False):
    """
    Preprocesses the configuration to resolve internal references

    - resolve $.x paths
    - resolves $self.y
    - resolves $.x[0]
    - resolves $.x[$.y[0]
    """
    if root is None:
        root = config

    if isinstance(config, list):
        return [parse_references(v, root, this, validate) for v in config]

    if isinstance(config, tuple):
        return (parse_references(v, root, this, validate) for v in config)

    if isinstance(config, dict):
        for k, v in config.items():
            if validate:
                # error if reserved keywords are being used
                if k in _reserved_keys:
                    raise ValueError(
                        f"'{k}' is a reserved name and cannot be used in machinable configuration keys"
                    )
                if k in _used_keys:
                    msg(
                        f"The configuration key '{k}' is a build-in mapping name and should not be used"
                        f"in machinable configuration keys as any dot-notation will become ambiguous.",
                        level="warning",
                        color="fail",
                    )

            config[k] = parse_references(v, root, this, validate)
        return config

    if isinstance(config, str) and (
        config.startswith("$.") or config.startswith("$self.")
    ):
        return parse_reference(config, root, this)

    return config


def auto_scope(key, scope):
    if key is None:
        return None

    if scope is None:
        return key

    # ignore imports
    if key.startswith("+."):
        return key

    # skip if already scoped
    if key.startswith(scope + "."):
        return key

    return scope + "." + key


def parse_module_list(
    collection,
    scope,
    baseclass,
    modules=None,
    imports=None,
    reference=None,
    import_prefix=None,
    auto_alias=None,
):
    if collection is None:
        return {}

    if modules is None:
        modules = {}

    if imports is None:
        imports = {}

    if "@" not in modules:
        modules["@"] = {}

    if isinstance(collection, (list, tuple)):
        for element in collection:
            modules.update(
                parse_module_list(
                    element,
                    scope,
                    baseclass,
                    modules,
                    imports,
                    reference,
                    import_prefix,
                    auto_alias,
                )
            )

        return modules

    for module, args in collection.items():

        # copy the default import prefix
        module_import_prefix = import_prefix

        # add scope to module
        module = auto_scope(module, scope)

        # parse alias
        if module.find("=") != -1:
            module, alias = module.split("=")
        else:
            alias = None

        # parse parent
        parent = None
        if module.find("^") != -1:
            module, parent = module.split("^")

            if parent.startswith("+."):
                inherited = get_or_fail(
                    imports,
                    parent.replace("+.", ""),
                    error="Dependency '^+.{}'  not found. Did you register it under '+'?",
                )

                if "_mixins_" in inherited["args"]:
                    inherited["args"]["_mixins_"] = [
                        {
                            "name": m["name"],
                            "vendor": parent.replace("+.", "").split(".")[0],
                        }
                        for m in parse_mixins(inherited["args"]["_mixins_"])
                    ]
            else:
                inherited = get_or_fail(
                    modules,
                    auto_scope(parent, scope),
                    error="Parent module '^{}' of " + module + " does not exist.",
                )

            # inherit the parent's config
            args = update_dict(inherited["args"], args, copy=True)
            # if no module name specified, use same as parent
            if module == "":
                module = inherited["module"]
                # disable import path as import modules already incorporate import prefixes
                module_import_prefix = None

        # module specification
        if args is None:
            args = {}

        # if import module, check if existing
        if module.startswith("+."):
            import_config = get_or_fail(
                imports,
                module.replace("+.", ""),
                error="Dependency '+.{}' not found. Did you register it under '+'?",
            )

            d = import_config["args"].copy()
            if import_config["args"].get("_mixins_"):
                d["_mixins_"] = [
                    {
                        "name": m["name"],
                        "vendor": module.replace("+.", "").split(".")[0],
                    }
                    for m in parse_mixins(d["_mixins_"])
                ]

            # inherit the imports' config
            args = update_dict(d, args, copy=True)

        # parse references
        if reference is not None:
            args = parse_references(args, root=reference, this=args, validate=True)

        # get actual module import path location
        module_import = module.replace("+.", "vendor.")
        if module_import_prefix:
            module_import = module_import_prefix + "." + module_import

        if module in modules:
            cls = modules[module]["class"]
        else:
            cls = ModuleClass(module_import, baseclass=baseclass)

        modules[module] = {"module": module_import, "class": cls, "args": args}

        # add alias lookup identity
        modules["@"][module] = module

        # add module alias
        scoped_alias = auto_scope(alias, scope)
        if scoped_alias is not None and scoped_alias != module:
            if not is_valid_variable_name(alias):
                raise ValueError(
                    f"Alias '{alias}' of '{module}' is not a valid Python variable name"
                )
            if alias in modules:
                raise ValueError(f"Alias '{alias}' of '{module}' is ambiguous")
            modules["@"][scoped_alias] = module
            modules[scoped_alias] = modules[module]

        # automatic alias
        aliased = module
        if auto_alias:
            prefix = auto_alias + "."
            if aliased.startswith(prefix):
                aliased = aliased[len(prefix) :]
                if aliased not in modules:
                    modules["@"][aliased] = module
                    modules[aliased] = modules[module]

        # make sure mixin have valid attribute name if not aliased
        if baseclass.__name__ == "Mixin" and not is_valid_variable_name(module):
            if not (is_valid_variable_name(alias) or is_valid_variable_name(aliased)):
                raise ValueError(
                    f"Mixin '{module}' has to be a valid Python variable name, please provide an alias"
                    f" e.g. '{module}=my_alias'"
                )

    return modules



import datetime
import json
import traceback

from ..utils.dicts import serialize


def msg(text, level="info", color=None):
    color_f = {
        "header": "\033[95m",
        "blue": "\033[94m",
        "green": "\033[92m",
        "yellow": "\033[93m",
        "fail": "\033[91m",
    }
    end_f = "\033[0m"

    if isinstance(color, str):
        text = color_f[color.lower()] + text + end_f

    print(text)


def str_to_time(date_time_str):
    if date_time_str is False or date_time_str is None:
        return None

    try:
        return datetime.datetime.strptime(date_time_str, "%Y-%m-%d %H:%M:%S.%f")
    except ValueError:
        return datetime.datetime.strptime(date_time_str, "%Y-%m-%d %H:%M:%S")


def prettydict(dict_like, sort_keys=False):
    try:
        return json.dumps(dict_like, indent=4, default=serialize, sort_keys=sort_keys)
    except TypeError:
        return str(dict_like)


def exception_to_str(ex):
    return "".join(
        traceback.format_exception(etype=type(ex), value=ex, tb=ex.__traceback__)
    )



import pytest

import machinable as ml


def test_ray_engine():
    t = ml.Experiment().components("thenode", "thechildren").repeat(2)
    ml.execute(t, engine="ray", project="./test_project")


@pytest.mark.last
def test_ray_engine_tune():
    import ray
    from ray import tune

    ray.init(ignore_reinit_error=True)
    ml.execute(
        ml.Experiment()
        .components("tunemodel")
        .tune(stop={"acc": 0.5}, config={"lr": tune.grid_search([0.001, 0.01])}),
        engine="ray",
        storage="./_test_data/tune",
        project="./test_project",
    )



import ast
import copy
import importlib

from ..core.exceptions import ExecutionException
from ..utils.dicts import update_dict
from ..utils.formatting import exception_to_str, msg
from ..utils.traits import Jsonable

_register = {
    "native": "machinable.engine.native_engine",
    "ray": "machinable.engine.ray_engine",
    "detached": "machinable.engine.detached_engine",
    "remote": "machinable.engine.remote_engine",
    "dry": "machinable.engine.dry_engine",
}


class Engine(Jsonable):
    @staticmethod
    def register(engine, name=None):
        if name is None:
            name = engine.__name__
        _register[name] = engine

    @classmethod
    def create(cls, args):
        if isinstance(args, Engine):
            return args

        if isinstance(args, dict):
            args = copy.deepcopy(args)

        if isinstance(args, str):
            args = {"type": args}

        if args is None:
            args = {"type": "native"}

        engine = args.pop("type")

        arg = []
        if engine.find(":") != -1:
            engine, version = engine.split(":", maxsplit=1)
            try:
                options = ast.literal_eval(version)
            except ValueError:
                options = version
            if isinstance(options, dict):
                args = update_dict(args, options)
            elif isinstance(options, (list, tuple)):
                arg.extend(options)
            else:
                arg.append(options)

        try:
            if isinstance(_register[engine], str):
                engine_module = importlib.import_module(_register[engine])
                class_name_snake = _register[engine].split(".")[-1]
                class_name = "".join(
                    p.capitalize() for p in class_name_snake.split("_")
                )
                _register[engine] = getattr(engine_module, class_name)
        except KeyError:
            raise ValueError(f"Unknown engine: {engine}.")
        except ImportError as ex:
            raise ValueError(f"Engine import failed: {exception_to_str(ex)}")
        except AttributeError:
            raise ValueError(f"Engine could not be found.")

        return _register[engine](*arg, **args)

    @classmethod
    def unserialize(cls, serialized):
        return cls.create(serialized)

    def submit(self, execution):
        """Retrieves an execution instance for execution

        Must call execution.set_result() with result and
        return the execution instance

        # Arguments
        execution: machinable.Execution

        machinable.Execution object
        """
        for (
            index,
            (execution_type, component, components, storage, resources, args, kwargs,),
        ) in enumerate(execution.schedule.iterate(execution.storage)):
            result = self.process(
                execution_type, component, components, storage, resources, args, kwargs
            )
            execution.set_result(result)

        return execution

    def process(self, execution_type, *args, **kwargs):
        return getattr(self, execution_type)(*args, **kwargs)

    def storage_middleware(self, storage):
        return storage

    def log(self, text, level="info"):
        msg("Engine: " + text, level, color="blue")

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "machinable.Engine"

    def execute(
        self,
        component,
        components=None,
        storage=None,
        resources=None,
        args=None,
        kwargs=None,
    ):
        return ExecutionException(
            reason="unsupported",
            message="The engine does not support execution operations",
        )

    def tune(
        self,
        component,
        components=None,
        storage=None,
        resources=None,
        args=None,
        kwargs=None,
    ):
        return ExecutionException(
            reason="unsupported",
            message="The engine does not support tuning operations",
        )



import machinable as ml


def test_project_export(helpers):
    path = helpers.tmp_directory("export")
    ml.Execution(
        experiment=ml.Experiment().components("nodes.observations", "export_model"),
        project="./test_project",
    ).export(path=path + "/test",)



import copy
import os

import ray
from ray.exceptions import RayActorError

from ..core.component import FunctionalComponent
from ..core.exceptions import ExecutionException
from ..utils.dicts import update_dict
from ..utils.formatting import exception_to_str
from .engine import Engine


class RayEngine(Engine):
    def serialize(self):
        return {"type": "ray"}

    def submit(self, execution):
        if not ray.is_initialized():
            ray.init()

        for arguments in execution.schedule.iterate(execution.storage):
            try:
                result = self.process(*arguments)
                if isinstance(result, ray.ObjectID):
                    result = ray.get(result)
                execution.set_result(result)
            except RayActorError as ex:
                result = ExecutionException(
                    reason="exception",
                    message=f"The following exception occurred: {ex}\n{exception_to_str(ex)}",
                )
                execution.set_result(result)

        return execution

    def execute(
        self,
        component,
        components=None,
        storage=None,
        resources=None,
        args=None,
        kwargs=None,
    ):
        if isinstance(component["class"], FunctionalComponent):
            nd = ray.remote(resources=resources)(FunctionalComponent).remote(
                component["class"].function, component["args"], component["flags"],
            )
        else:
            nd = ray.remote(resources=resources)(component["class"]).remote(
                component["args"], component["flags"]
            )

        # destroy events class as independent instances will be recreated in the local workers
        storage["events"] = None

        return nd.dispatch.remote(components, storage, nd)

    def tune(
        self,
        component,
        components=None,
        store=None,
        resources=None,
        args=None,
        kwargs=None,
    ):
        # local import to enable Ray driver use without ray[tune] dependencies
        import ray.tune
        from ray.tune.trainable import Trainable
        from ray.tune.result import DONE

        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}

        if "events" in store:
            del store["events"]

        class MachinableTuningComponent(Trainable):
            def _setup(self, config):
                # apply config updates
                components_update = [{} for _ in range(len(components))]
                if "node" in config:
                    node_update = config["node"]
                    if "components" in config:
                        components_update = config["children_patch"]
                else:
                    node_update = config

                node_args = copy.deepcopy(component["args"])
                node_args = update_dict(node_args, node_update)
                node_flags = copy.deepcopy(component["flags"])
                node_flags["TUNING"] = True

                components_config = copy.deepcopy(components)
                for i in range(len(components)):
                    components_config[i]["args"] = update_dict(
                        components[i]["args"], components_update[i]
                    )

                storage_config = copy.deepcopy(store)
                storage_config["url"] = "../../../"
                storage_config["allow_overwrites"] = True

                self.node = component["class"](node_args, node_flags)
                self.components, self.store = self.node.dispatch(
                    components_config, storage_config, self, lifecycle=False
                )
                self.node.create(self.components, self.store)
                self.node.execute()
                self._node_execution_iterations = -1

            def _train(self):
                if hasattr(self.node.on_execute_iteration, "_deactivated"):
                    raise NotImplementedError(
                        "on_execute_iteration has to be implemented to allow for step-wise tuning"
                    )

                self._node_execution_iterations += 1
                self.node.flags.ITERATION = self._node_execution_iterations
                self.node.on_before_execute_iteration(self._node_execution_iterations)
                try:
                    callback = self.node.on_execute_iteration(
                        self._node_execution_iterations
                    )
                    if (
                        self.node.on_after_execute_iteration(
                            self._node_execution_iterations
                        )
                        is not False
                    ):
                        # trigger records.save() automatically
                        if (
                            self.node.store.has_records()
                            and not self.node.store.record.empty()
                        ):
                            self.store.record[
                                "_iteration"
                            ] = self._node_execution_iterations
                            self.store.record.save()
                except (KeyboardInterrupt, StopIteration):
                    callback = StopIteration

                if len(self.node.record.history) > self._node_execution_iterations:
                    data = self.node.record.history[-1].copy()
                else:
                    data = {}

                data[DONE] = callback is StopIteration

                return data

            def _save(self, tmp_checkpoint_dir):
                return self.node.save_checkpoint(tmp_checkpoint_dir)

            def _restore(self, checkpoint):
                self.node.restore_checkpoint(checkpoint)

            def _stop(self):
                self.node.destroy()

        if "url" not in store:
            store["url"] = "~/ray_results"

        if store["url"].find("://") != -1:
            fs_prefix, local_dir = store["url"].split("://")
            if fs_prefix != "osfs":
                raise ValueError(
                    "Store has to be local; use upload_dir or sync options of Ray tune to copy to remote."
                )
        else:
            local_dir = store["url"]

        kwargs["name"] = os.path.join(store["group"], store["uid"])
        kwargs["resources_per_trial"] = resources
        kwargs["local_dir"] = local_dir

        output = ray.tune.run(MachinableTuningComponent, *args, **kwargs)

        return output



import copy
import datetime
import json
import pickle
from collections import OrderedDict

import pendulum

from ..utils.formatting import msg, prettydict, serialize


class Record:
    """Tabular record writer

    ::: tip
    Becomes available as ``self.store.record`` and ``self.record``
    :::

    # Arguments
    store: machinable.store.Observer, parent store instance
    config: dict, configuration options
    scope: String, optional destination name
    """

    def __init__(self, store, config=None, scope="default"):
        self.store = store
        self.config = config if config is not None else {}
        self.scope = scope
        self.history = []
        self._record = OrderedDict()
        # create records directory
        self.store.get_path("records", create=True)

    def write(self, key, value, fmt=None):
        """Writes a cell value

        Note that you can use array notation to write value as well, for example

        ```python
        self.record.write('loss', 0.1)
        # is equivalent to
        self.record['loss'] = 0.1
        ```

        # Arguments
        key: String, the column name
        value: Value to write
        fmt: String, optional formatter.
        """
        if fmt == "seconds":
            value = str(datetime.timedelta(seconds=value))

        self._record[key] = value

    def update(self, dict_like=None, **kwargs):
        """Update record values using a dictionary. Equivalent to dict's update method.

        # Arguments
        dict_like: dict, update values
        """
        self._record.update(dict_like, **kwargs)

    def empty(self):
        """Whether the record writer is empty (len(self.record) == 0)
        """
        return len(self._record) == 0

    def timing(self, mode="iteration", timestamp=None, return_type="seconds"):
        """Get timing statistics about the records

        # Arguments
        mode: String, 'iteration', 'avg' or 'total' that determines whether the last iteration, the average iteration
              or the total time is collected
        timestamp: Mixed, end of the timespan; if None, datetime.now() is used
        return_type: String, specifying the return format 'seconds', 'words', or 'period' (pendulum.Period object)

        Returns: See return_type
        """
        if timestamp is None:
            timestamp = datetime.datetime.now()

        if "on_execute_start" not in self.store.statistics:
            self.store.statistics["on_execute_start"] = datetime.datetime.now()

        start = self.store.statistics["on_execute_start"]

        if mode == "iteration":
            if len(self.history) > 0:
                start = self.history[-1]["_timestamp"]
            elapsed = pendulum.instance(start).diff(timestamp)
        elif mode == "total":
            if len(self.history) > 0:
                start = self.history[0]["_timestamp"]
            elapsed = pendulum.instance(start).diff(timestamp)
        elif mode == "avg":
            if len(self.history) > 0:
                start = self.history[0]["_timestamp"]
                elapsed = pendulum.instance(start).diff(timestamp) / len(self.history)
            else:
                elapsed = pendulum.instance(start).diff(timestamp)
        else:
            raise ValueError(
                f"Invalid mode: '{mode}'; must be 'iteration', 'avg' or 'total'."
            )

        if return_type == "seconds":
            seconds = elapsed.in_seconds()
            if seconds == 0:
                return 1e-15  # avoids division by zero errors
            return seconds
        elif return_type == "words":
            return elapsed.in_words()
        elif return_type == "period":
            return elapsed
        else:
            raise ValueError(
                f"Invalid return_type: '{return_type}'; must be 'seconds', 'words' or 'period'."
            )

    def save(self, echo=False, force=False):
        """Save the current row

        # Arguments
        echo: Boolean, if True the written row will be printed to the terminal
        force: Boolean, if True the row will be written even if it contains no columns

        # Returns
        The row data
        """
        data = copy.deepcopy(self._record)
        self._record = OrderedDict()

        # don't save if there are no records
        if len(data) == 0 and not force:
            return data

        # meta-data
        data["_timestamp"] = datetime.datetime.now()

        iteration_time = self.timing(
            "iteration", timestamp=data["_timestamp"], return_type="period"
        )
        data["_seconds_taken"] = iteration_time.in_seconds()
        data["_time_taken"] = iteration_time.in_words()

        total_time = self.timing(
            "total", timestamp=data["_timestamp"], return_type="period"
        )
        data["_total_seconds_taken"] = total_time.in_seconds()
        data["_total_time_taken"] = total_time.in_words()

        self.history.append(data)

        if self.scope == "default":
            if hasattr(self.store, "events"):
                self.store.events.trigger("store.record.on_save", data)

        # json
        with self.store.get_stream(f"records/{self.scope}.json", "w") as f:
            json.dump(self.history, f, ensure_ascii=False, default=serialize)

        # pickle
        with self.store.get_stream(f"records/{self.scope}.p", "wb") as f:
            pickle.dump(self.history, f)

        if self.scope == "default":
            if hasattr(self.store, "events"):
                self.store.events.trigger("store.on_change", "record.save")

        if echo:
            msg(self.store.directory())
            msg(prettydict(data, sort_keys=True))

        return data

    def __len__(self):
        return len(self._record)

    def __delitem__(self, key):
        del self._record[key]

    def __getitem__(self, key):
        return self._record[key]

    def __setitem__(self, key, value):
        self._record[key] = value



import copy
import csv
import datetime
import importlib
import inspect
import json
import logging
import os
import random
import re
import sys
from collections import OrderedDict
from importlib import reload
from inspect import getattr_static
from keyword import iskeyword
from typing import Dict, List, Mapping, Optional, Type, Union

try:
    import cPickle as pickle
except ImportError:
    import pickle

# !include machinable.utils.host


class ConfigMap(dict):
    def __init__(self, dict_like, *args, **kwargs):
        for key, value in dict_like.items():
            if hasattr(value, "keys"):
                value = ConfigMap(value)
            self[key] = value

    def toDict(self, *args, **kwargs):
        return self

    __delattr__ = dict.__delitem__
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


def config_map(d=None):
    return ConfigMap(d)


# !include machinable.utils.formatting.msg


# !include machinable.utils.dicts.serialize


# !include machinable.utils.utils.is_valid_variable_name


# !include machinable.utils.dicts.update_dict


# !include machinable.utils.utils.apply_seed


# !include machinable.utils.formatting.prettydict


# !include machinable.store.record.Record


# !include machinable.store.log.Log


class FileSystem:
    def __init__(self, directory):
        self.directory = directory

    def makedirs(self, path, recreate=True):
        os.makedirs(os.path.join(self.directory, path), exist_ok=True)

    def makedir(self, path, recreate=True):
        return self.makedirs(path, recreate)

    def isfile(self, file):
        return os.path.isfile(os.path.join(self.directory, file))

    def exists(self, path):
        return os.path.exists(os.path.join(self.directory, path))

    def open(self, path, *args, **kwargs):
        path = os.path.join(self.directory, path)
        return open(path, *args, **kwargs)


def open_fs(storage, *args, **kwargs):
    return FileSystem(storage)


# !include machinable.store.store.OutputRedirection


# !include machinable.store.store.Store


# !include machinable.core.component.set_alias


# !include machinable.core.component.inject_components


# !include machinable.config.parser.ModuleClass


# !include machinable.config.mapping.ConfigMethod


# !include machinable.core.component.bind_config_methods


# !include machinable.config.parser.parse_mixins


# !include machinable.core.component.MixinInstance


Mixin = object
# !include machinable.Mixin


# !include machinable.core.component.ComponentState


# !include machinable.core.component.Component



import os
import sys
from shlex import quote

from ..core.settings import get_settings
from .engine import Engine


class DetachedEngine(Engine):
    def __init__(self, engine=None, using="tmux", close=True):
        if using not in ["tmux"]:
            raise ValueError(f"Invalid detached mode: {using}")
        self.engine = Engine.create(engine)
        self.using = using
        self.close = close

    def submit(self, execution):
        name = "machinable-experiment-" + execution.experiment_id

        url = os.path.join(
            execution.storage.get("url", "mem://"), execution.experiment_id
        )
        engine = self.engine.to_json().replace('"', '\\"')
        project = execution.project.to_json().replace('"', '\\"')
        code = f"""
        import machinable as ml
        e = ml.Execution.from_storage('{url}')
        e.set_engine(ml.Engine.from_json('{engine}'))
        e.set_storage({execution.storage})
        e.set_project(ml.Project.from_json('{project}'))
        e.submit()
        """.replace(
            "\n        ", ";"
        )[
            1:-1
        ]
        command = f'{sys.executable or "python"} -c "{code}"'
        if self.close:
            command += "; exit"
        self.shell(command, name=name)

        return execution

    def storage_middleware(self, storage):
        if not storage.get("url", "mem://").startswith("mem://"):
            return storage

        # use temporary directory
        storage["url"] = get_settings()["tmp_directory"]

        # todo$: how to clean up?

        return storage

    def serialize(self):
        return {
            "type": "detached",
            "engine": self.engine.serialize(),
            "using": self.using,
            "close": self.close,
        }

    @classmethod
    def unserialize(cls, serialized):
        serialized["engine"] = Engine.unserialize(serialized["engine"])
        return cls.create(serialized)

    def __repr__(self):
        return f"Detached({repr(self.engine)}, using={self.using})"

    def shell(self, command, name):
        return getattr(self, self.using + "_shell")(command, name)

    # using

    def tmux_shell(self, command, name):
        return os.system(
            f"tmux new-session -d -s {name}; tmux send-keys {quote(command)} Enter"
        )



import datetime
import time

from machinable import Component


class Timings(Component):
    def on_create(self):
        q = self.record.timing("avg")
        assert q > 0

        self.record["test"] = 1
        self.record.save(echo=True)

        self.record["test"] = 2
        self.record.save(echo=True)

        assert self.record.timing("avg") > 0



from machinable import Component


class MyBaseModel(Component):
    def on_create(self):
        self.epoch = None

    def on_execute(self):
        for epoch in range(5):
            self.epoch = epoch
            print(epoch, self.config.learning_rate)

    def config_learning_rate(self, base=0.1):
        if not self.epoch:
            return base  # default learning rate

        return base * self.epoch



import hashlib
import os
import random
import string
from keyword import iskeyword


def get_file_hash(filepath):
    """Returns a hash value of a file

    # Arguments
    filepath: Absolute filepath

    # Returns
    md5 hexdigest of file content
    """
    if not os.path.isfile(filepath):
        return None
    algorithm = hashlib.md5()
    with open(filepath, "rb") as f:
        file_content = f.read()
        algorithm.update(file_content)

    return algorithm.hexdigest()


def apply_seed(seed=None):
    """Applies a global random seed to the application state.

    In particular, the method seeds Python's random module and numpy, tensorflow and pytorch packages if available.

    Arguments:
    seed: Int|None, the random seed. Use None to unset seeding
    """
    if not isinstance(seed, int):
        return False

    random.seed(seed)

    try:
        import numpy as np

        np.random.seed(seed)
    except ImportError:
        pass

    try:
        import tensorflow as tf

        try:
            tf.random.set_seed(seed)
        except AttributeError:
            tf.compat.v1.set_random_seed(seed)

    except ImportError:
        pass

    try:
        import torch

        torch.manual_seed(seed)
    except ImportError:
        pass

    return True


def generate_seed(random_state=None):
    """Generates a seed from a random state

    # Arguments
    random_state: Random state or None

    Returns:
    int32 seed value
    """
    if random_state is None or isinstance(random_state, int):
        random_state = random.Random(random_state)

    return random_state.randint(0, 2 ** 31 - 1)


def is_valid_variable_name(name):
    if not isinstance(name, str):
        return False
    return name.isidentifier() and not iskeyword(name)


def random_str(length, random_state=None):
    if random_state is None or isinstance(random_state, int):
        random_state = random.Random(random_state)

    return "".join(
        random_state.choice(
            string.ascii_uppercase + string.ascii_lowercase + string.digits
        )
        for _ in range(length)
    )


def generate_uid(k=1, random_state=None):
    if random_state is None or isinstance(random_state, int):
        random_state = random.Random(random_state)

    if not isinstance(k, int):
        raise ValueError(f"k has to be integer, {type(k)} given")

    return [random_str(length=12, random_state=random_state) for _ in range(k)]



import copy

import yaml

from ..experiment import ExperimentComponent
from ..utils.dicts import update_dict
from .parser import ModuleClass, parse_mixins


def collect_updates(version):
    collection = []
    for update in version:
        arguments = update["arguments"]

        if not isinstance(arguments, tuple):
            arguments = (arguments,)

        collection.extend(arguments)

    return collection


class ConfigInterface:
    def __init__(self, parsed_config, version=None, default_class=None):
        self.data = parsed_config
        if version is None:
            version = []
        self.version = version
        self.default_class = default_class

    def _get_version(self, name, config, current=None):
        # from yaml string
        if not name.startswith("~") and not (
            name.startswith("_") and name.endswith("_")
        ):
            return yaml.load(name)

        # from mixin
        if name.startswith("_") and name.endswith("_"):
            try:
                return copy.deepcopy(self.data["mixins"][name[1:-1]]["args"])
            except KeyError:
                raise KeyError(
                    f"Mixin '{name}' not available.\n"
                    f"Did you register it under 'mixins'?\n"
                )

        # from local version
        if name in config:
            version = copy.deepcopy(config[name])
            if not current:
                return version
            else:
                # nested lookup
                try:
                    return update_dict(copy.deepcopy(current[name]), version)
                except KeyError:
                    return version

        if current:
            # nested lookup
            try:
                return copy.deepcopy(current[name])
            except KeyError:
                # ignore non-existing nested lookups
                return {}

        raise KeyError(
            f"Version '{name}' could not be found.\n"
            f"Available versions: {[f for f in config.keys() if f.startswith('~')]}\n"
        )

    def get_component(self, name, version=None, flags=None):
        if name is None:
            return None

        if name not in self.data["components"]:
            raise ValueError(
                f"Component '{name}' not found. Is it registered in the machinable.yaml?"
            )

        if self.default_class is not None:
            self.data["components"][name]["class"] = self.default_class

        # load lazy modules
        if isinstance(self.data["components"][name]["class"], ModuleClass):
            self.data["components"][name]["class"] = self.data["components"][name][
                "class"
            ].load(instantiate=False, default=self.default_class)

        # de-reference
        config = copy.deepcopy(self.data["components"][name])

        # flags
        config["flags"] = flags if flags is not None else {}

        # name
        config["name"] = name

        # introspection
        config["flags"]["NAME"] = name
        config["flags"]["MODULE"] = config["module"]

        # carry over global evaluation config
        if "_evaluate" not in config["args"]:
            config["args"]["_evaluate"] = self.data["_evaluate"]

        # un-alias
        origin = self.data["components"]["@"][name]

        # mixins
        for i, mixin in enumerate(parse_mixins(config["args"].get("_mixins_"))):
            mixin_info = {"name": mixin["name"]}

            if mixin["name"].startswith("+."):
                raise AttributeError(
                    f"Mixin import '{mixin['name']}' is ambiguous. "
                    f"Please register it in the mixins section and use a unique alias."
                )

            if origin.startswith("+."):
                vendor = origin.split(".")[1]

                # consider import mixins
                try:
                    mixin_args = copy.deepcopy(
                        self.data["imports"]["mixins"][vendor + "." + mixin["name"]]
                    )
                except KeyError:
                    raise KeyError(
                        f"Dependent mixin '{vendor + '.' + mixin['name']}' of '{origin}' not found."
                    )

                # un-alias
                mixin_info["origin"] = (
                    "+."
                    + vendor
                    + "."
                    + self.data["imports"]["mixins"][vendor + ".@"].get(mixin["name"])
                )
            else:
                if "vendor" in mixin:
                    try:
                        mixin_args = copy.deepcopy(
                            self.data["imports"]["mixins"][
                                mixin["vendor"] + "." + mixin["name"]
                            ]
                        )
                        mixin_info["origin"] = (
                            "+."
                            + mixin["vendor"]
                            + "."
                            + self.data["imports"]["mixins"][mixin["vendor"] + ".@"][
                                mixin["name"]
                            ]
                        )
                    except KeyError:
                        raise KeyError(
                            f"Dependent mixin '{mixin['vendor'] + '.' + mixin['name']}' not found."
                        )
                else:
                    try:
                        mixin_args = copy.deepcopy(self.data["mixins"][mixin["name"]])
                        mixin_info["origin"] = self.data["mixins"]["@"].get(
                            mixin["name"]
                        )
                    except KeyError:
                        raise KeyError(
                            f"Mixin '{mixin['name']}' not found. Is it registered under 'mixins'?"
                        )

            # the mixin config can be overwritten by the local config so we update the mixin arg and write it back
            config["args"] = update_dict(mixin_args["args"], config["args"], copy=True)

            # write information
            config["args"]["_mixins_"][i] = mixin_info

        # parse updates

        # merge local update to global updates
        config["versions"] = []
        versions = copy.deepcopy(self.version)
        if version is not None:
            versions.append({"arguments": {"components": copy.deepcopy(version)}})

        version = {}
        for updates in collect_updates(versions):
            # select components/node subsection
            update = updates.get("components", None)
            if update is None:
                continue

            if not isinstance(update, tuple):
                update = (update,)

            for k in update:
                # load arguments from machinable.yaml
                if isinstance(k, str):
                    config["versions"].append(k)
                    k = self._get_version(k, config["args"], version)

                # merge with version
                version = update_dict(version, k)

        config["args"] = update_dict(config["args"], version)

        # remove unused versions
        config["args"] = {
            k: v
            for k, v in config["args"].items()
            if not k.startswith("~") or k in config["versions"]
        }

        return config

    def get(self, component, components=None):
        component = ExperimentComponent.create(component)
        node_config = self.get_component(
            component.name, component.version, component.flags
        )

        if components is None:
            return node_config

        components_config = []
        for c in components:
            subcomponent = ExperimentComponent.create(c)
            component_config = self.get_component(
                subcomponent.name, subcomponent.version, subcomponent.flags
            )
            if component_config is not None:
                components_config.append(component_config)

        return node_config, components_config



import datetime
import os

import pendulum
from fs.errors import CreateFailed as CreateFailedException
from orator import DatabaseManager, Model

from machinable.v1.history import get_history

from ..observations.orm.migrations import run as migrate_database
from ..observations.orm.models import ObservationModel, StorageModel, TaskModel
from ..observations.orm.query_builder import QueryBuilder
from .collections import Collection, ObservationCollection
from .views import ObservationView, TaskView


class Observations:
    """Interface to manage machinable observations

    The interface is read-only. machinable will never modify any of the collected observations.

    # Examples

    ```python
    import machinable as ml
    mlo = ml.Observations('~/observations')
    mlo.find()
    ```

    # Arguments
    write: String, URL of write location.
    database: String, optional path to SQLite database used for indexing.
        Defaults to non-persistent in-memory database.
    """

    def __init__(self, storage=None, database=None):
        self._cache = {}
        self._registry = set()
        self.database = None
        self._init(database or ":memory:")  # todo: use settings
        # check write locations in database
        for s in StorageModel.all():
            if not s.available():
                s.delete()
                continue

            # push unindexed write locations to lazy load registry
            if not s.indexed:
                self._registry.add(s.url)

        if storage is not None:
            self.add(storage)

    def _init(self, database=":memory:"):
        database = os.path.expanduser(database)

        self.database = DatabaseManager(
            {"observations": {"driver": "sqlite", "database": database}}
        )
        Model.set_connection_resolver(self.database)

        if os.path.isfile(database):
            # do not re-initialise if file based database
            return

        # create and migrate database
        if database != ":memory:":
            os.makedirs(os.path.dirname(database), exist_ok=True)
            open(database, "w").close()

        migrate_database(self.database)

        return True

    @property
    def storages(self):
        self._lazy_index()
        return list(StorageModel.all(["url"]).map(lambda storage: storage.url))

    @property
    def available_storages(self):
        return [e for e in get_history().available() if e not in self.storages]

    def reset(self, database=":memory:"):
        """Resets the interface and deletes the database index

        # Arguments
        database: String, optional location of the SQLite database
        """
        if self.database is not None:
            self.database.disconnect()
            self.database = None

        self._registry = set()
        self._cache = {}

        if os.path.isfile(database):
            os.remove(database)

        return self._init(database)

    def refresh(self, storage=None, or_add=True):
        """Reloads all observations

        # Arguments
        write: String, optional URL of write that should be reloaded. If None, all registered storages
            will be reloaded
        or_add: If True, non-existing write location will be added otherwise ignored.
        """
        if isinstance(storage, str):
            model = StorageModel.whereUrl(storage).first()
            if model is None:
                if or_add:
                    return self.add(storage)
                else:
                    return False
            model.index()
        else:
            # remove database and repopulate as probably faster
            storages = StorageModel.all()
            self.reset()
            for storage in storages:
                self.add(storage.url)

        self._cache = {}
        return True

    def add(self, storage, lazy=False):
        """Load a write location

        # Arguments
        write: String, URL to index
        lazy: Boolean, whether location should only be loaded on demand. Defaults to False.
        """
        if isinstance(storage, (list, tuple)):
            return [self.add(s) for s in storage]

        if lazy:
            # push lazy load registry
            self._registry.add(storage)

        model = StorageModel.whereUrl(storage).first()

        if model:
            return

        model = StorageModel.create(url=StorageModel.get_url(storage))
        if not lazy:
            try:
                model.index()
            except CreateFailedException:
                raise FileNotFoundError(f"{storage} does not exists or is not readable")

        return model.url

    def remove(self, storage):
        """Unloads a write location.

        Note that this does NOT remove any data, the location can be reloaded at any time.
        """
        model = StorageModel.whereUrl(storage).first()

        if not model:
            return

        model.delete()
        return True

    def _lazy_index(self):
        while self._registry:
            self.add(self._registry.pop(), lazy=False)

    # query builder

    @property
    def query(self):
        """Returns a [query builder](#observationsquerybuilder) to build customized queries"""
        self._lazy_index()
        return ObservationsQueryBuilder(self._cache)

    # standard queries

    def find(self, observation_id=None, rerun=False):
        """Finds an observation

        # Arguments
        observation_id: String, observation ID. If None, all available observations will be returned.
            Alternatively, a 6-digit experiment ID or a 9-digit execution ID can be used to retrieve observations.
        rerun: Integer, optional number of rerun if observation has been reproduced multiple times

        # Returns
        Instance or collection of machinable.Observation
        """
        if rerun is False:
            if observation_id is None:
                return self.query.get()

            if isinstance(observation_id, str):
                if len(observation_id) == 6:
                    return self.find_by_task(observation_id)

                if len(observation_id) == 9:
                    return self.find_by_execution(observation_id)

        if isinstance(observation_id, (list, tuple)):
            return self.find_many(observation_id, rerun)
        return (
            self.query.where("path", observation_id)
            .where_has("task", lambda q: q.where("rerun", int(rerun)))
            .first()
        )

    def find_all(self):
        """Returns a collection of all available observations

        # Returns
        Instance or collection of machinable.Observation
        """
        return self.query.get()

    def find_many(self, observations, rerun=False):
        """Finds many observations

        # Arguments
        observations: List of observation ID strings.
        rerun: Integer, optional number of rerun if observations have been reproduced multiple times

        # Returns
        Instance or collection of machinable.Observation
        """
        return (
            self.query.where_in("path", observations)
            .where_has("task", lambda q: q.where("rerun", int(rerun)))
            .get()
        )

    def find_by_task(self, task_id):
        """Finds all observations of a experiment

        # Arguments
        task_id: String, experiment ID. Alternatively, a list of experiment ID can be used.

        # Returns
        Instance or collection of machinable.Observation
        """
        if isinstance(task_id, (list, tuple)):
            return self.query.where_task_in("task_id", task_id).get()

        return self.query.latest().where_task(task_id).get()

    def find_by_most_recent_task(self, criterion="started"):
        """Finds all observations of the most recent experiment

        # Arguments
        criterion: String, the field by which the tasks should be ordered in descending order.

        # Returns
        Instance or collection of machinable.Observation or None if no tasks are available
        """
        task = TaskModel.order_by(criterion, "desc").first()
        if task is None:
            return None
        return self.find_by_task(task.task_id)

    def find_by_execution(self, execution_id):
        """Finds observations of a given execution

        # Arguments
        execution_id: String|Experiment|Observation that specifies the execution

        # Returns
        Instance or collection of machinable.Observation
        """

        def _to_id(execution):
            if isinstance(execution, Collection):
                execution = execution.first()

            if isinstance(execution, (TaskView, ObservationView)):
                execution = execution.execution_id

            if not isinstance(execution, str):
                raise ValueError(f"'{execution}' is not a valid execution ID.")

            return execution

        if isinstance(execution_id, (list, tuple)):
            execution_id = list(map(_to_id, execution_id))
            return self.query.where_task_in("execution_id", execution_id).get()

        return (
            self.query.latest()
            .where_task("execution_id", "=", _to_id(execution_id))
            .get()
        )

    def find_by_task_name(self, name):
        """Finds observations of tasks with a given name

        # Arguments
        name: String, the search name

        # Returns
        Instance or collection of machinable.Observation
        """
        return self.query.latest().where_task("name", "=", name).get()

    def find_by_node_component(self, node):
        """Finds observations that were produced by a given node components

        # Arguments
        node: String, the search name

        # Returns
        Instance or collection of machinable.Observation
        """
        return self.query.latest().where("node", "=", node).get()

    def find_by_child_component(self, child):
        """Finds observations that were produced by a given child components

        # Arguments
        child: String, the search name

        # Returns
        Instance or collection of machinable.Observation
        """
        return self.query.latest().where(f"FIND_IN_SET('{child}')", "<>", child).get()

    def find_by_storage(self, storage):
        """Finds observations of a given write

        # Arguments
        write: String, the search name

        # Returns
        Instance or collection of machinable.Observation
        """
        if isinstance(storage, (list, tuple)):
            return self.query.where_storage_in(storage).get()

        return self.query.latest().where_storage(storage).get()


class ObservationsQueryBuilder(QueryBuilder):
    def __init__(self, cache=None):
        self._cache = cache
        self._query = ObservationModel.with_("task.storage")

    def _field_mapping(self, field):
        if field == "id":
            field = "path"

        return field

    def _collection(self, models):
        return ObservationCollection(
            models.transform(lambda model: ObservationView(model, self._cache))
        )

    # query operations

    def first(self):
        """Returns the first observation of the specified query
        """
        model = self._query.first()
        if model is None:
            return None
        return ObservationView(model, self._cache)

    def get(self):
        """Returns the observations of the specified query
        """
        models = self._query.get()
        return self._collection(models)

    # aggregated

    def finished(self, time=None, operator="<"):
        """"""
        if time is None:
            time = datetime.datetime.now()
        self._query.where_not_null("finished").where("finished", operator, time)
        return self

    def not_finished(self):
        """"""
        self._query.where_is_null("finished")
        return self

    def alive(self):
        """"""
        self._query.where("heartbeat", ">", pendulum.now().subtract(seconds=30))
        return self

    def tuning(self):
        """"""
        return self.where_task("tuning", "=", 1)

    def not_tuning(self):
        """"""
        return self.where_task("tuning", "=", 0)

    def has_code_backup(self):
        """"""
        return self.where_task("code_backup", "=", True)

    def has_no_code_backup(self):
        """"""
        return self.where_task("code_backup", "=", False)

    def rerun(self, number):
        """"""
        return self.where_task("rerun", "=", number)

    # query mutators

    def latest(self, field="started"):
        """"""
        return self.order_by(field, "asc")

    def oldest(self, field="started"):
        """"""
        return self.order_by(field, "desc")

    # where_storage

    def where_storage(self, url, boolean="and"):
        """"""
        self._query = self._query.where_has(
            "write", lambda q: q.where("url", "=", url, boolean)
        )

        return self

    def or_where_storage(self, url):
        """"""
        return self.where_storage(url, "or")

    def where_storage_in(self, values, boolean="and", negate=False):
        """"""
        self._query = self._query.where_has(
            "write", lambda q: q.where_in("url", values, boolean, negate)
        )

        return self

    def or_where_storage_in(self, values):
        """"""
        return self.where_storage_in(values, "or")

    def where_not_storage_in(self, values, boolean="and"):
        """"""
        return self.where_storage_in(values, boolean, True)

    def or_where_storage_not_in(self, values):
        """"""
        return self.where_storage_in(values, "or")

    # where_task

    def where_task(self, field, operator="=", value=None, boolean="and"):
        """"""
        if field == "id":
            field = "task_id"

        if value is None:
            value = field
            field = "task_id"

        self._query = self._query.where_has(
            "task", lambda q: q.where(field, operator, value, boolean)
        )

        return self

    def or_where_task(self, field, operator="=", value=None):
        """"""
        return self.where_task(field, operator, value, "or")

    def where_task_between(self, field, values, boolean="and", negate=False):
        """"""
        if field == "id":
            field = "task_id"

        self._query = self._query.where_has(
            "task", lambda q: q.where_between(field, values, boolean, negate)
        )

        return self

    def or_where_task_between(self, field, values):
        """"""
        return self.where_task_between(field, values, "or")

    def where_not_task_between(self, field, values, boolean="and"):
        """"""
        return self.where_task_between(field, values, boolean, True)

    def or_where_not_task_between(self, field, values):
        """"""
        return self.where_not_task_between(field, values, "or")

    def where_task_in(self, field, values, boolean="and", negate=False):
        """"""
        if field == "id":
            field = "task_id"

        self._query = self._query.where_has(
            "task", lambda q: q.where_in(field, values, boolean, negate)
        )

        return self

    def or_where_task_in(self, field, values):
        """"""
        return self.where_task_in(field, values, "or")

    def where_not_task_in(self, field, values, boolean="and"):
        """"""
        return self.where_task_in(field, values, boolean, True)

    def or_where_task_not_in(self, field, values):
        """"""
        return self.where_not_task_in(field, values, "or")



import machinable as ml


def test_detached_engine():
    t = ml.Experiment().component("thenode", {"v": 1})
    # smoke test
    execution = ml.Execution(
        t, engine=ml.engine.Detached(using="tmux"), project="./test_project",
    ).submit()



import os

from ariadne import load_schema_from_path, make_executable_schema
from ariadne.asgi import GraphQL
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
from starlette.routing import Route

from .filesystem import storage_resolver
from .graphql import mutation, query, scalar_types, subscription

dir_path = os.path.dirname(os.path.realpath(__file__))
schema = make_executable_schema(
    load_schema_from_path(os.path.join(dir_path, "graphql/schema/")),
    [*scalar_types, query, mutation, subscription],
)


graphql = GraphQL(schema, keepalive=True)

routes = [
    Route("/filesystem/storage/{url:path}/{filename:path}", endpoint=storage_resolver,)
]

middleware = [
    Middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_methods=["*"],
        allow_headers=["*"],
        allow_credentials=True,
    )
]

server = Starlette(routes=routes, middleware=middleware)
server.mount("/", graphql)



# optimization.py

from machinable import Component


class DummyOptimization(Component):
    def on_create(self):
        print(
            "Creating the optimization model with the following configuration: ",
            self.config,
        )

    def on_execute(self):
        for i in range(3):
            print("Training step", i)



from machinable import Component


class TuneModel(Component):
    def on_execute(self):
        pass

    def on_execute_iteration(self, iteration: int):
        self.record["acc"] = iteration * 0.1

        if iteration > 10:
            return StopIteration



import pytest

from machinable.config.parser import parse_mixins, parse_reference


def test_parse_reference():
    root = {
        "test": 1,
        "me": {
            "nested": 2,
            "deep": {"structure": 3, "with": {"nested": 4, "to": "check"}},
        },
        "key": "nested",
        "hello": {"world": "with"},
        "list": ["great", "stuff", "beautiful"],
    }
    this = root["me"]

    assert parse_reference("$self.nested", root, this) == 2
    assert parse_reference("$.test", root, this) == 1

    assert parse_reference("$.me[$.key]", root, this) == 2

    assert parse_reference("$.list[$self.nested]", root, this) == "beautiful"
    assert parse_reference("$.list[$.me[nested]]", root, this) == "beautiful"
    assert parse_reference("$.list[$.me[$.key]]", root, this) == "beautiful"

    assert parse_reference("$self.deep[$.hello[world]][$.key]", root, this) == 4
    assert parse_reference("$self.deep[$.hello[world]].to", root, this) == "check"


def test_parse_mixins():
    assert parse_mixins(None) == []

    with pytest.raises(ValueError):
        parse_mixins({})

    # converts str into list
    assert parse_mixins("mixin")[0]["name"] == "mixin"

    # attributes
    assert parse_mixins("+.import.project")[0]["attribute"] == "_import_project_"
    assert parse_mixins("module.path")[0]["attribute"] == "_module_path_"
    assert parse_mixins("invalid$name")[0]["valid_attribute"] is False



import copy
import json
import os
from shlex import quote

import sh

from ..core.exceptions import ExecutionException
from .engine import Engine


class RemoteEngine(Engine):
    def __init__(self, engine=None, host=None, directory=None, python="python"):
        if host is None:
            raise ValueError("You have to provide a remote host")
        if directory is None:
            raise ValueError("You have to provide a remote directory")
        self.engine = Engine.create(engine)
        self.host = host
        self.directory = directory
        self.python = python
        self.shell = sh.ssh.bake(self.host)

    def serialize(self):
        return {
            "type": "remote",
            "engine": self.engine.serialize(),
            "host": self.host,
            "directory": self.directory,
            "python": self.python,
        }

    @classmethod
    def unserialize(cls, serialized):
        serialized["engine"] = Engine.unserialize(serialized["engine"])
        return cls.create(serialized)

    def submit(self, execution):
        if execution.storage.get("url", "mem://").startswith("mem://"):
            raise ValueError("Remote engine does not support temporary file systems")
        self.log(
            f"Rsyncing project {execution.project.directory_path} -> {self.host}:{self.directory}"
        )
        path = execution.project.directory_path
        if path[-1] == "/":
            path = path[:-1]
        sh.rsync(
            "-az", path, f"{self.host}:{self.directory}",
        )
        self.log("Rsync complete. Submitting execution.")

        # todo$: handle local URLs and ssh URL correctly
        #  i.e. mirror local one to remote, remove ssh prefix if same as self.host

        url = os.path.join(execution.storage["url"], execution.experiment_id)
        engine = self.engine.to_json().replace('"', '\\"')
        remote_project = copy.deepcopy(execution.project.options)
        remote_project["directory"] = self.directory
        project = json.dumps(remote_project).replace('"', '\\"')
        code = f"""
        import machinable as ml
        e = ml.Execution.from_storage('{url}')
        e.set_engine(ml.Engine.from_json('{engine}'))
        e.set_storage({execution.storage})
        e.set_project(ml.Project.from_json('{project}'))
        e.submit()
        """.replace(
            "\n        ", ";"
        )[
            1:-1
        ]
        command = f'cd {self.directory}; {self.python} -c "{code}"'
        try:
            self.shell(quote(command))
            execution.set_result(True)
        except sh.ErrorReturnCode as ex:
            execution.set_result(
                ExecutionException(ex.stderr.decode("utf-8"), reason="engine_failure")
            )
            self.log(f"Execution failed: {str(ex)}", level="error")

        return execution

    def __repr__(self):
        return f"Remote({repr(self.engine)})"



import copy
import datetime
import io
import json
import os
import sys

from fs import open_fs

from ..utils.dicts import serialize, update_dict
from .log import Log
from .record import Record

try:
    import cPickle as pickle
except ImportError:
    import pickle


class OutputRedirection:
    def __init__(self, stream_type, mode, file_open, file_name=None):
        if stream_type not in ["stdout", "stderr"]:
            raise ValueError(f"Invalid stream type: {stream_type}")

        self._file_stream = None
        self.stream_type = stream_type
        self.mode = mode
        self.file_open = file_open
        self.file_name = file_name or stream_type + ".log"
        self.sys_stream = getattr(sys, stream_type)

        # capture output from other consumers of the underlying file descriptor
        if self.mode != "DISCARD":
            try:
                output_file_descriptor = os.dup(self.sys_stream.fileno())
                os.dup2(self.file_stream.fileno(), output_file_descriptor)
            except (AttributeError, IOError, io.UnsupportedOperation):
                pass

    @property
    def file_stream(self):
        if self._file_stream is None:
            self._file_stream = self.file_open(self.file_name, "a", buffering=1)

        return self._file_stream

    @property
    def streams(self):
        if self.mode == "DISCARD":
            return []

        if self.mode == "FILE_ONLY":
            return [self.file_stream]

        return [self.file_stream, self.sys_stream]

    def write(self, message):
        for i, stream in enumerate(self.streams):
            try:
                stream.write(message)
            except IOError:
                if i == 0:
                    # close corrupt file stream
                    self.close_file_stream()

    def close_file_stream(self):
        try:
            self._file_stream.close()
        except (IOError, AttributeError):
            pass
        finally:
            self._file_stream = None

    # forward attributes to standard sys stream

    def __getattr__(self, item):
        return getattr(self.sys_stream, item)

    @classmethod
    def apply(cls, mode, file_open, file_name=None):
        if mode not in ["DISABLED", "FILE_ONLY", "SYS_AND_FILE", "DISCARD"]:
            raise ValueError(f"Invalid output redirection mode: {mode}")

        if mode == "DISABLED":
            return

        sys.stdout.flush()
        sys.stderr.flush()
        sys.stdout = cls("stdout", mode, file_open, file_name)
        sys.stderr = cls("stderr", mode, file_open, file_name)

    @classmethod
    def revert(cls):
        if isinstance(sys.stdout, cls):
            sys.stdout.close_file_stream()
            sys.stdout = sys.stdout.sys_stream
        if isinstance(sys.stderr, cls):
            sys.stderr.close_file_stream()
            sys.stderr = sys.stderr.sys_stream


class Store:
    """Store interface

    ::: tip
    Becomes available as ``self.store``
    :::

    # Arguments
    config: dict, configuration options
    heartbeat: Integer, seconds between two heartbeat events
    """

    def __init__(self, config=None, status=True):
        if isinstance(config, dict):
            config = copy.deepcopy(config)

        if isinstance(config, str):
            config = {"url": config}

        # default settings
        self.config = update_dict(
            {
                "url": "mem://",
                "log": {},
                "records": {},
                "group": "",
                "heartbeat": 15,
                "output_redirection": "SYS_AND_FILE",  # DISABLED, FILE_ONLY, SYS_AND_FILE, DISCARD
                "code_backup": None,
            },
            config,
            copy=True,
        )
        self.statistics = {}

        if "://" not in self.config["url"]:
            self.config["url"] = "osfs://" + self.config["url"]

        # event system
        from ..core.events import Events

        if isinstance(self.config.get("events", None), Events):
            self.events = self.config["events"]
        else:
            self.events = Events()
        self.events.heartbeats(seconds=self.config["heartbeat"])

        self.filesystem = open_fs(self.config["url"], create=True)
        self.filesystem.makedirs(
            self.get_path(), recreate=self.config.get("allow_overwrites", False)
        )

        self._record_writers = {}
        self._log = None
        self._store = {}

        # status
        self._status = dict()
        self._status["started"] = str(datetime.datetime.now())
        self._status["finished"] = False
        if status:
            self.refresh_status()
            self.events.on("heartbeat", self.refresh_status)

        if not self.config["url"].startswith("mem://"):
            OutputRedirection.apply(
                self.config["output_redirection"], self.get_stream, "output.log"
            )

    def destroy(self):
        """Destroys the store instance"""
        # disable heartbeat
        if "events" in self.config:
            self.events.heartbeats(None)
        # write finished status (not triggered in the case of an unhandled exception)
        self._status["finished"] = str(datetime.datetime.now())
        self.refresh_status()

        OutputRedirection.revert()

    def directory(self):
        """Returns the write directory"""
        return os.path.join(
            self.config["url"], self.config["group"], self.config.get("uid", "")
        )

    def local_directory(self):
        """Returns the local filesystem path, or False if non-local

        # Returns
        Local filesystem path, or False if non-local
        """
        if not self.config["url"].startswith("osfs://"):
            return False

        return self.config["url"].split("osfs://")[-1]

    def refresh_status(self):
        """Updates the status.json file with a heartbeat at the current time
        """
        if not self.config.get("uid", None):
            return
        self._status["heartbeat"] = str(datetime.datetime.now())
        self.write("status.json", self._status, overwrite=True, _meta=True)

    def get_record_writer(self, scope):
        """Creates or returns an instance of a record writer

        # Arguments
        scope: Name of the record writer

        # Returns
        machinable.store.record.Record
        """
        if scope not in self._record_writers:
            self._record_writers[scope] = Record(
                store=self, config=self.config["records"], scope=scope
            )

        return self._record_writers[scope]

    @property
    def stored(self):
        return self._store

    @property
    def record_writers(self):
        return self._record_writers

    @property
    def record(self):
        """Record interface

        # Returns
        machinable.store.record.Record
        """
        return self.get_record_writer("default")

    @property
    def log(self):
        """Log interface

        # Returns
        machinable.store.log.Log
        """
        if self._log is None:
            self._log = Log(observer=self, config=self.config["log"])

        return self._log

    def has_records(self, scope="default"):
        """Determines whether record writer exists

        # Arguments
        scope: String, name of the record writer. Defaults to 'default'

        # Returns
        True if records have been written
        """
        return scope in self._record_writers

    def has_logs(self):
        """Determines whether log writer exists

        # Returns
        True if log writer exists
        """
        return self._log is not None

    def store(self, name, data, overwrite=False, _meta=False):
        # todo: remove v1 deprecation
        return self.write(name, data, overwrite, _meta)

    def write(self, name, data, overwrite=False, _meta=False):
        """Stores a data object

        # Arguments
        name: String, name identifier.
            You can provide an extension to instruct machinable to write the data in its own file and not as part
            of a dictionary with other stored values.
            Supported formats are .json (JSON), .npy (numpy), .p (pickle), .txt (txt)
        data: The data object
        overwrite: Boolean, if True existing values will be overwritten
        """
        mode = "w" if overwrite else "a"
        path = os.path.dirname(name)
        name = os.path.basename(name)
        _, ext = os.path.splitext(name)

        if not _meta:
            path = "store/" + path
            self.filesystem.makedir(self.get_path(path), recreate=True)
        filepath = os.path.join(path, name)

        # todo: check overwrite for files

        if ext == "":
            # automatic handling
            if name in self._store and not overwrite:
                raise ValueError(
                    f"'{name}' already exist. "
                    f"Use overwrite=True if you intent to overwrite existing data"
                )
            self._store[name] = data
            with self.get_stream("store.json", "w") as f:
                f.write(json.dumps(self._store, ensure_ascii=False, default=serialize))
        elif ext == ".json":
            # json
            with self.get_stream(filepath, mode) as f:
                f.write(json.dumps(data, ensure_ascii=False, default=serialize))
        elif ext == ".npy":
            import numpy as np

            if "b" not in mode:
                mode += "b"
            # numpy
            with self.get_stream(filepath, mode) as f:
                np.save(f, data)

        elif ext == ".p":
            if "b" not in mode:
                mode += "b"
            with self.get_stream(filepath, mode) as f:
                pickle.dump(data, f)

        elif ext == ".txt":
            with self.get_stream(filepath, mode) as f:
                f.write(data)
        else:
            raise ValueError(
                f"Invalid format: '{ext}'. "
                f"Supported formats are .json (JSON), .npy (numpy), .p (pickle), .txt (txt)"
            )

        if "events" in self.config:
            self.events.trigger(
                "store.on_change", "store.save", {"name": name, "data": data}
            )

    def get_stream(self, path, mode="r", *args, **kwargs):
        """Returns a file stream on the store write

        # Arguments
        path: Relative file path
        mode: Python file mode
        *args, **kwargs: Optional arguments passed into stream open()
        """
        return self.filesystem.open(self.get_path(path), mode, *args, **kwargs)

    def get_path(self, append="", create=False):
        """Returns the write path

        # Arguments
        append: String, optional postfix that is appended to the path
        create: Boolean, if True path is being created if not existing
        """
        path = os.path.join(self.config["group"], self.config.get("uid", ""), append)

        if create:
            self.filesystem.makedirs(path, recreate=True)

        return path



import json


class Jsonable:
    def to_json(self, stringify=True):
        serialized = self.serialize()
        if stringify:
            serialized = json.dumps(serialized)
        return serialized

    def serialize(self):
        raise NotImplementedError

    @classmethod
    def from_json(cls, serialized):
        if isinstance(serialized, str):
            serialized = json.loads(serialized)
        return cls.unserialize(serialized)

    @classmethod
    def unserialize(cls, serialized):
        raise NotImplementedError

    def clone(self):
        return self.__class__.from_json(self.to_json())



import machinable as ml
from machinable.experiment.parser import parse_experiment


def seeding_test(spec):
    assert (
        len(set([e.flags["GLOBAL_SEED"] for e, _, _ in spec])) == 1
    )  # repeats share same GLOBAL_SEED
    assert len(set([e.flags["SEED"] for e, _, _ in spec])) > 1  # has unique SEED


def test_experiment_parser():
    # repeat behaviour
    t = ml.Experiment().components("test")
    assert len(list(parse_experiment(t.specification))) == 1
    t = ml.Experiment().components("test").repeat(1)
    assert len(list(parse_experiment(t.specification))) == 1

    t = ml.Experiment().components("test").repeat(5)
    spec = list(parse_experiment(t.specification))
    assert len(spec) == 5
    seeding_test(spec)

    t = ml.Experiment().components(("test", "~v"), ("test", "~v")).repeat(3)
    spec = list(parse_experiment(t.specification))
    assert len(spec) == 3
    seeding_test(spec)

    t = ml.Experiment().components("test").repeat(3).split(2)
    spec = list(parse_experiment(t.specification))
    assert len(spec) == 6
    seeding_test(spec)

    t = ml.Experiment().components(("test", [{"a": i} for i in range(3)]))
    spec = list(parse_experiment(t.specification))
    assert len(spec) == 3
    seeding_test(spec)

    t = ml.Experiment().components(("test", [{"a": i} for i in range(3)])).repeat(3)
    spec = list(parse_experiment(t.specification))
    assert len(spec) == 9
    seeding_test(spec)

    t = ml.Experiment().components("test", ("test", [{"a": i} for i in range(3)]))
    spec = list(parse_experiment(t.specification))
    assert len(spec) == 3
    seeding_test(spec)

    t = ml.Experiment().components(
        "test",
        [
            ("test", [{"a": i} for i in range(3)]),
            ("test", [{"a": i} for i in range(3)]),
        ],
    )
    spec = list(parse_experiment(t.specification))
    assert len(spec) == 9
    seeding_test(spec)

    t = ml.Experiment().components("test", ("test", [{"sub": lr} for lr in range(5)]))
    spec = list(parse_experiment(t.specification))
    assert len(spec) == 5
    seeding_test(spec)



import webbrowser

import click
import uvicorn

from ..server.server import server


@click.command()
@click.option("--host", default="127.0.0.1", help="Host")
@click.option("--port", default=5000, help="Port")
@click.option("--log-level", default="info", help="Log level")
def server(host, port, log_level):
    uvicorn.run(server, host=host, port=port, log_level=log_level)


@click.command()
def app():
    host = "127.0.0.1"
    webbrowser.open("http://app.machinable.org/", new=0, autoraise=True)
    # todo: find free port and encode the endpoint via ?get request or #
    # todo: secure endpoint with token request
    uvicorn.run(server, host=host, port=5000, log_level="warning")



import os

from machinable.core.settings import get_settings

from ..utils.formatting import msg


def fetch_directory(source, target):
    if not os.path.isdir(source):
        raise ValueError(f"{source} is not a directory")

    # we use symlink rather than copy for performance and consistency
    os.symlink(source, os.path.join(target), target_is_directory=True)

    return True


def fetch_git(source, target):
    raise FileNotFoundError(
        f"Dependency missing. Please 'git clone {source}' into {target}"
    )


def fetch_import(source, target):
    if source is None:
        return False

    # git
    if source.startswith("git+"):
        return fetch_git(source[4:], target)

    # default: directory
    return fetch_directory(source, target)


def fetch_imports(project, config=None):
    if config is None:
        config = project.get_config()["+"]
    rc = get_settings()
    top_level_vendor = project.get_root().vendor_path
    os.makedirs(top_level_vendor, exist_ok=True)

    imports = []
    for dependency in config:
        if isinstance(dependency, str):
            dependency = {dependency: None}

        for name, args in dependency.items():
            imports.append(name)

            # source config
            if not isinstance(args, str):
                args = None

            # give local .machinablerc priority over project config
            source = rc["imports"].get(name, args)
            if isinstance(source, str):
                source = os.path.expanduser(source)

            # local target folder
            os.makedirs(os.path.join(project.directory_path, "vendor"), exist_ok=True)
            target = os.path.join(project.directory_path, "vendor", name)

            # protect against invalid symlinks
            if os.path.islink(target) and not os.path.exists(target):
                os.unlink(target)

            # skip if already existing
            if os.path.exists(target):
                break

            # top-level target folder
            top_level = os.path.join(top_level_vendor, name)

            # protect against invalid symlinks
            if os.path.islink(top_level) and not os.path.exists(top_level):
                os.unlink(top_level)

            # fetch import to the top-level if it does not exist
            if not os.path.exists(top_level):
                msg(f"Fetching '+.{name}' to {top_level}")
                if not fetch_import(source, top_level):
                    raise FileNotFoundError(
                        f"Could not fetch '+.{name}'. Please place it into {top_level}"
                    )

            # symlink from top-level to local target if not identical
            if top_level != target:
                os.symlink(top_level, target, target_is_directory=True)

            break

    return imports



from machinable import Component


class ExceptionsComponent(Component):
    def on_create(self):
        raise ValueError("Hihihi ...")

    def on_execute(self):
        raise TypeError("Sorry, but no")

    def on_execute_iteration(self, iteration: int):
        raise StopIteration("Don't let anything stop you")



import machinable as ml


def test_core_execute_decorator():
    @ml.execute
    def run(node, components, store):
        assert node.config.alpha == 0
        store.log.info("Custom training with learning_rate=" + str(node.config.a))
        assert components[0].config.alpha == 0

    t = ml.Experiment().components("thenode", "thechildren")
    run(t, seed=1, project="./test_project")



from .collection import Collection


class RecordsCollection(Collection):
    def as_dataframe(self):
        """Returns collection as Pandas dataframe
        """
        import pandas

        data = (
            {k: [row[k] for row in self._items] for k in self._items[0].keys()}
            if len(self._items) > 0
            else {}
        )
        return pandas.DataFrame.from_dict(data)



import os

_history = None


def get_history(reload=False, file=None):
    global _history
    if _history is None or reload:
        _history = History(file)

    return _history


class History:
    def __init__(self, file=None):
        if file is None:
            file = "history"
        os.makedirs(os.path.expanduser("~/.machinable"), exist_ok=True)
        self.horizon = 15
        self.url = os.path.expanduser(os.path.join("~/.machinable", file))
        self.data = []
        self._load()

    def _load(self):
        if not os.path.isfile(self.url):
            self.data = []
            return

        with open(self.url, "r") as f:
            for line in f:
                self.data.append(line)

    def _save(self):
        if len(self.data) == 0:
            try:
                os.remove(self.url)
            except OSError:
                pass
            return

        with open(self.url, "w+") as f:
            for element in self.data:
                f.write(element + "\n")

    def available(self):
        return [e for e in self.data if os.path.isfile(e)]

    def add(self, storage):
        if not isinstance(storage, str):
            return False

        storage = storage.strip()
        if len(storage) == 0:
            return False

        if storage in self.data:
            self.data.remove(storage)
        self.data.append(storage)
        if len(self.data) > self.horizon:
            self.data = self.data[-self.horizon :]
        self._save()

    def reset(self):
        self.data = []
        self._save()

    def __len__(self):
        return len(self.data)



import os

import yaml

from machinable.utils.dicts import update_dict

_settings = None


def get_settings(reload=False, file="~/.machinable/settings.yaml"):
    global _settings
    if _settings is None or reload:
        try:
            with open(os.path.expanduser(file), "r") as f:
                _settings = yaml.load(f)
        except FileNotFoundError:
            _settings = {}
        except yaml.parser.ParserError as e:
            _settings = {"_errors": f"Invalid configuration file: {e}"}

        # defaults
        _settings = update_dict(
            {
                "cache": {"imports": False},
                "imports": {},
                "database": {
                    "default": "sqlite",
                    "sqlite": {"driver": "sqlite", "database": ":memory:"},
                },
                "tmp_directory": "userdata://machinable:machinable/tmp",
            },
            _settings,
        )

    return _settings



import click

import pkg_resources

from .server import app, server
from .vendor.commands import vendor


@click.group(invoke_without_command=True)
@click.pass_context
@click.option("--version", "-v", is_flag=True, help="Prints the version number")
def cli(ctx, version):
    if ctx.invoked_subcommand is not None:
        return

    if version:
        click.echo(pkg_resources.require("machinable")[0].version)
        return

    click.echo(ctx.get_help())


cli.add_command(vendor)
cli.add_command(server)
cli.add_command(app)



from multiprocessing import Pool

from .engine import Engine


class NativeEngine(Engine):
    def __init__(
        self, processes=1,
    ):
        self.processes = processes

    def __repr__(self):
        return "Native"

    def serialize(self):
        return {"type": "native", "processes": self.processes}

    def submit(self, execution):
        if self.processes <= 1:
            # standard execution
            return super(NativeEngine, self).submit(execution)

        pool = Pool(processes=self.processes)
        results = pool.starmap(
            self.process,
            [arguments for arguments in execution.schedule.iterate(execution.storage)],
        )

        for result in results:
            execution.set_result(result)

        return execution

    def execute(
        self,
        component,
        components=None,
        storage=None,
        resources=None,
        args=None,
        kwargs=None,
    ):
        if resources is not None:
            self.log(
                "Resource specification are currently not supported. Use Ray Engine instead.",
                level="warn",
            )

        nd = component["class"](component["args"], component["flags"])
        return nd.dispatch(components, storage)



from machinable import Component


class ConfMethods(Component):
    def on_create(self):
        assert self.config.method == "test"
        assert self.config.argmethod == "world"
        assert self.config.nested.method == "test"

    def config_hello(self):
        return "test"

    def config_arghello(self, arg):
        return arg



# language=yaml
alexnet = """
learning_rate: 0.1
data:
  name: mnist
"""

task = ml.Task().components(("optimization", alexnet))

# the above is equivalent to:

task = ml.Task().components(
    ("optimization", {"learning_rate": 0.1, "data": {"name": "mnist"}})
)



import copy
import datetime
import os
from typing import Any, Callable, Union

import fs
import yaml

from machinable.utils.host import get_host_info

from ..config.interface import ConfigInterface
from ..core.exceptions import ExecutionException
from ..engine import Engine
from ..execution.schedule import Schedule
from ..experiment.experiment import Experiment
from ..experiment.parser import parse_experiment
from ..project import Project
from ..project.export import Export
from ..storage.directory import Directory as StorageDirectory
from ..store import Store
from ..utils.formatting import exception_to_str, msg
from ..utils.traits import Jsonable
from ..utils.utils import generate_uid
from .identifiers import (
    decode_experiment_id,
    encode_experiment_id,
    generate_experiment_id,
)


def to_color(experiment_id):
    return "".join(
        [encode_experiment_id(decode_experiment_id(c) % 16) for c in experiment_id]
    )


class Execution(Jsonable):
    def __init__(
        self,
        experiment: Union[Experiment, Any],
        storage: Union[dict, str] = None,
        engine: Union[Engine, str, dict, None] = None,
        project: Union[Project, Callable, str, dict] = None,
        seed: Union[int, None, str] = None,
    ):
        self._uid = None
        self._uid_cache = None
        self.function = None

        self.experiment = None
        self.storage = None
        self.engine = None
        self.project = None
        self.seed = None

        self.timestamp = None
        self.schedule = Schedule()
        self.store = None
        self.code_backup = None
        self.code_version = None
        self.started_at = None

        if callable(experiment):
            # decorator use
            if None not in (storage, engine, project, seed):
                raise ValueError(
                    "Execution decorator takes no arguments; "
                    "call the decorated function with arguments instead."
                )
            self.function = experiment
            return

        self.set_experiment(experiment)
        self.set_storage(storage)
        self.set_engine(engine)
        self.set_project(project)
        self.set_seed(seed)

    def __call__(self, experiment, storage=None, engine=None, project=None, seed=None):
        self.set_experiment(experiment)
        self.set_storage(storage)
        self.set_engine(engine)
        self.set_project(project)
        self.set_seed(seed)
        if self.function is not None:
            # decorator invocation
            self.project.default_component = self.function
            return self

    def set_experiment(self, experiment):
        self.experiment = Experiment.create(experiment)

        return self

    def set_storage(self, storage):
        if isinstance(storage, dict):
            storage = copy.deepcopy(storage)
        elif isinstance(storage, str):
            storage = {"url": storage}
        else:
            storage = {}

        # code backup
        self.code_backup = False
        if storage.get("code_backup", None) is not False:
            if not (
                # do not backup on mem:// filesystem unless explictly set to True
                storage.get("code_backup", None) is None
                and storage.get("url", "mem://").startswith("mem://")
            ):
                self.code_backup = True

        self.storage = storage

        return self

    def set_engine(self, engine):
        self.engine = Engine.create(engine)

        return self

    def set_project(self, project):
        self.project = Project.create(project)

        return self

    def set_seed(self, seed):
        if isinstance(seed, str):
            seed = decode_experiment_id(seed, or_fail=True)
        elif isinstance(seed, int):
            seed = generate_experiment_id(random_state=seed, with_encoding=False)
        else:
            seed = generate_experiment_id(with_encoding=False)
        self.seed = seed

        return self

    def set_schedule(self, schedule=None):
        if schedule is not None:
            self.schedule = schedule
            return self

        # build schedule
        self.schedule = Schedule()

        config = ConfigInterface(
            self.project.parse_config(),
            self.experiment.specification["version"],
            default_class=self.project.default_component,
        )

        for index, (node, components, resources) in enumerate(
            parse_experiment(self.experiment.specification, seed=self.seed)
        ):
            node_config, components_config = config.get(node, components)

            if "tune" in self.experiment.specification:
                self.schedule.add_tune(
                    component=node_config,
                    components=components_config,
                    resources=resources,
                    **self.experiment.specification["tune"]["arguments"],
                )
            else:
                self.schedule.add_execute(
                    component=node_config,
                    components=components_config,
                    resources=resources,
                )

        return self

    @property
    def uid(self):
        if self.seed is None:
            return []

        if self._uid is None or self._uid_cache != self.seed:
            self._uid = generate_uid(k=len(self.schedule), random_state=self.seed)
            self._uid_cache = self.seed

        return self._uid

    def is_submitted(self):
        try:
            storage = fs.open_fs(self.storage["url"], create=False)
            return storage.isdir(self.experiment_id)
        except (
            AttributeError,
            KeyError,
            TypeError,
            ValueError,
            fs.errors.ResourceNotFound,
            fs.errors.CreateFailed,
        ):
            pass
        return False

    def submit(self):
        if not self.is_submitted():
            if len(self.schedule) == 0:
                self.set_schedule()
            now = datetime.datetime.now()
            self.timestamp = now.timestamp()
            self.started_at = str(now)
            self.storage["group"] = self.experiment_id
            self.store = Store(
                config=self.engine.storage_middleware(self.storage), status=False
            )

            if self.code_backup is not False:
                self.project.backup_source_code(opener=self.store.get_stream)

            if self.code_version is not False:
                self.code_version = self.project.get_code_version()

            self.store.write("execution.json", self.serialize(), _meta=True)
            self.store.write("host.json", get_host_info(), _meta=True)

        return self.engine.submit(self)

    def set_result(self, result, index=None):
        if index is None:
            index = len(self.schedule) - 1
        if isinstance(result, ExecutionException):
            self.engine.log(
                f"{self.uid[index]} of experiment {self.experiment_id} failed "
                f"({index + 1}/{len(self.schedule)}). "
                f"{exception_to_str(result)}",
                level="error",
            )
        else:
            self.engine.log(
                f"{self.uid[index]} of experiment {self.experiment_id} finished "
                f"({index + 1}/{len(self.schedule)}). ",
                level="info",
            )
        self.schedule.set_result(result, index)

    @classmethod
    def from_storage(cls, url):
        storage = StorageDirectory(url)
        return cls.from_json(storage.load_file("execution.json"))

    def export(self, path=None, overwrite=False):
        """Exports the execution

        Converts the execution into a plain Python project
        that can be executed without machinable.

        ::: warning
        This feature may not work reliably in all circumstances and project use cases
        :::

        # Arguments
        path: String, directory where exported execution will be stored. If None defaults to 'exports' in the
            current working directory
        overwrite: Boolean, whether to overwrite an existing export.
        """
        if path is None:
            path = os.path.join(os.getcwd(), "exports")

        if len(self.schedule) == 0:
            self.set_schedule()

        for (
            index,
            (execution_type, component, components, resources, args, kwargs),
        ) in enumerate(self.schedule):
            if execution_type != "execute":
                msg("Not supported", level="error", color="fail")
                continue

            # if more than one job, write into subdirectories
            if len(self.schedule) > 1:
                path = os.path.join(path, str(index))

            export_path = os.path.abspath(path)

            if os.path.exists(export_path) and not overwrite:
                raise FileExistsError(
                    f"Export directory '{export_path}' exists. To overwrite, set overwrite=True"
                )

            msg(f"\nExporting to {export_path}", color="yellow")

            export = Export(self.project.directory_path, export_path)

            # instantiate targets
            nd = component["class"](config=component["args"], flags=component["flags"])
            comps = [
                c["class"](config=c["args"], flags=c["flags"], node=nd)
                for c in components
            ]

            # export config
            export.write(
                "config.json",
                {
                    "component": {
                        "args": nd.config.toDict(evaluate=True),
                        "flags": nd.flags.toDict(evaluate=True),
                    },
                    "components": [
                        {
                            "args": comps[i].config.toDict(evaluate=True),
                            "flags": comps[i].flags.toDict(evaluate=True),
                            "class": components[i]["class"].__name__,
                        }
                        for i in range(len(comps))
                    ],
                    "storage": {"url": "./results"},
                },
                meta=True,
            )

            # export components and node
            export.module(component["class"])
            for c in components:
                export.module(c["class"])

            # export mixins
            mixins = component["args"].get("_mixins_", [])
            for c in components:
                mixins.extend(c["args"].get("_mixins_", []))
            for mixin in mixins:
                export.module(
                    mixin["origin"].replace("+.", "vendor."), from_module_path=True
                )

            export.write("__init__.py")

            export.machinable()

            export.entry_point(component, components)

            export.echo()

            msg(
                f"\nExporting finished. Run as 'cd {export_path} && python run.py'",
                color="yellow",
            )

    def serialize(self):
        return {
            "id": self.experiment_id,
            "seed": self.seed,
            "timestamp": self.timestamp,
            "code_backup": self.code_backup,
            "code_version": self.code_version,
            "started_at": self.started_at,
            "schedule": self.schedule.serialize(),
            "uid": self.uid,
        }

    @classmethod
    def unserialize(cls, serialized):
        if not isinstance(serialized, dict):
            raise ValueError(f"Invalid execution: {serialized}")
        execution = cls(None)
        execution.seed = serialized["seed"]
        execution.timestamp = serialized["timestamp"]
        execution.code_backup = serialized["code_backup"]
        execution.code_version = serialized["code_version"]
        execution.started_at = serialized["started_at"]
        execution.schedule = Schedule.unserialize(serialized["schedule"])
        return execution

    @property
    def experiment_id(self):
        return encode_experiment_id(self.seed)

    def summary(self):
        if len(self.schedule) == 0:
            self.set_schedule()

        msg(
            f"\nExecution: {self.experiment_id}\n----------", color="header",
        )
        msg(f"Storage: {self.storage.get('url', 'mem://')}", color="blue")
        msg(f"Engine: {repr(self.engine)}", color="blue")
        msg(f"Project: {repr(self.project)}", color="blue")

        for (
            index,
            (execution_type, component, components, resources, args, kwargs),
        ) in enumerate(self.schedule):
            # only print first and last one
            shortened = len(self.schedule) > 3 and 0 < index <= len(self.schedule) - 1
            msg(
                f"\n{self.uid[index]} ({index+1}/{len(self.schedule)})", color="header",
            )

            msg(f"\nComponent: {component['name']}", color="yellow")
            msg(f":: {component['flags']}")
            if len(component["versions"]) > 0:
                msg(f">> {', '.join(component['versions'])}", color="green")
            msg(yaml.dump(component["args"]), color="blue")

            for c in components:
                if c:
                    msg(f"\t {c['name']}", color="yellow")
                    msg(f"\t:: {c['flags']}")
                    if len(c["versions"]) > 0:
                        msg(f"\t>> {', '.join(c['versions'])}", color="green")
                    msg("\t" + yaml.dump(c["args"]), color="blue")

        return self
